/*************************************************************************
PROJECT:
   10-12 GHz solar radio telescope.

PROGRAM:
   CYCLOP14.C (V1.4)
   real-time 8-bit A/D converter measures acquisition.

AUTHOR:
   Nicolas Dupont-Bloch, France (Europe), september 2005.
   http://nicolas.dupontbloch.free.fr

REQUIREMENTS:
   Software:
   ---------
   C compiler (DJGPP, GCC...).
   Library Allegro.
   Linux, Dos(r) or Windows(r).
   
   Hardware:
   ---------
   Any PC with standard, even minimal graphics capabilities, and EPP parallel port.
   Hard disk with enough room for the log file, depending on the acquisition interval
   and the duration of the measures (27 bytes per measure, each 50ms up to 64s).
   The graphic resolution can be 320x200, 640x480, 800x600 (fine choice), or 1024x768.
   The spatial resolution depends on the antenna lobe and the measurement protocol
   (azimutal mount without drive, guided equatorial mount, fixed towards the Sun
   or X-Y scanning).
   The timeal resolution depends on the CPU (acquisition_interval).
   
   Radio-telescope:
   ----------------
   Antenna with 10-12GHz LNB, satellite-finder which signal to the buzzer
   is re-routed through a resistor divider to the A/D converter.
   8-bit A/D converter plugged into the parallel port in EPP mode.

THANKS:
   Bertrand Flouret, station astronomique Jansky, radio-observatoire de Nancay (France)
   for the idea and numerous explanations about radio-astronomy with such a device.
   http://astrosurf.com/saj
   ...and Allegro authors, tutorials and examples =:)

USE NOTES:
   1) In the BIOS at boot-time, set the parallel (LPT) port mode to EPP (Enhanced 
   Parallel Port),that is bi-directional mode.
   2) Run this program.
   3) Plug then switch the A/D converter on.

PARAMETERS:
   <prog_name> 320x200|640x480|800x600|1024x768 <log_file_name> [nowarning]
*************************************************************************/

#include "allegro.h"
#include <pc.h>
#include <dos.h>
#include <stdio.h>

#define  VERSION              "V1.4"
#define  K_ESCAPE             15131
#define  K_F1                 12032
#define  K_F2                 12288
#define  K_F3                 12544
#define  K_F4                 12800
#define  K_F5                 13056
#define  K_F6                 13312
#define  K_plus               22827
#define  K_minus              22573

#define  command_port         0x37A // parallel port (command)
#define  data_port            0x378 // parallel port (data)
#define  record_max           500
#define  converter_value_max  255   // max value from an 8-bit A/D converter.
#define  menu_line_max        7     // Height (nb of text lines) of top menu; remaining screen height is for graph.

struct record_t {
   struct   date record_date;
   struct   time record_time;
   int      solar_flux;
   };

// Variables used by real-time handler must be protected.
volatile struct record_t   record[record_max];
volatile FILE              *f_log = NULL;
volatile int               record_updated = FALSE;


void acquire_port(void) {
/*********************/
   outportb(command_port, 0x20);
   }

void release_port(void) {
/*********************/
   outportb(command_port, 0x0);
   }

void acquire_solar_flux(void) {
/***************************/
   char  message[64];   // For log file.
   int   i;
   
   // Keep and shift measures in array of records.
   for (i = 0; i < record_max; i++)
      record[i] = record[i+1];
   // Acquire new measure from A/D converter.
   record[record_max].solar_flux = inportb(data_port);
   gettime((struct time *) &record[record_max].record_time);
   getdate((struct date *) &record[record_max].record_date);
   // Write measure onto log file.
   if (f_log != NULL) {
      sprintf(message, "%04d-%02d-%02d %02d:%02d:%02d:%02d %03d", record[record_max].record_date.da_year, record[record_max].record_date.da_mon, record[record_max].record_date.da_day, record[record_max].record_time.ti_hour, record[record_max].record_time.ti_min, record[record_max].record_time.ti_sec, record[record_max].record_time.ti_hund, record[record_max].solar_flux);
      fputs(strcat(message, "\n"), (FILE *) f_log);
      }
   record_updated = TRUE; // Tell main loop that the record array has to be updated.
   }
END_OF_FUNCTION(acquire_solar_flux);


int open_log(const char * f_log_name) {
/***********************************/
   f_log = fopen(f_log_name, "a");
   if (f_log)
      return 1;
   else
      return 0;
   }

int close_log(void) {
/*****************/
   fclose((FILE *) f_log);
   f_log = NULL;
   return 1;
   }

int plotx(struct record_t rec, int x1, int x2, float hzoom) {
/*********************************************************/
   // x1 is graph_x1
   // x2 is graph_x2
   // hzoom is graph_hzoom
   int x;
   
   x = x1 + (hzoom * rec.solar_flux);
   if (x < x1)
      x = x1;
   else if (x > x2)
      x = x2;
   return x;
   }

int ploty(struct record_t rec, int y1, int y2, int vzoom, int i) {
/****************************************************************/
   // y1 is graph_y1
   // y2 is graph_y2
   // vzoom is graph_vzoom
   // i is the number of the record in its array.
   int y;
   
   y = y2 - (vzoom * (record_max - i));
   if (y < y1)
      y = y1+1;
   else if (y > y2)
      y = y2;
   return y;
   }

int main(int argc, char *argv[]) {
/******************************/
   #define     interval_max 10
   #define     interval_min 1       // 1st item (#0) of array is unused (1 to 10 is easier!).
   static int  acquisition_interval[interval_max+1] = // Timelapse (in ms) between acquisitions.
                  {0, 100, 250, 500, 1000, 2500, 5000, 10000, 15000, 30000, 64000};
   int         interval = 3;        // To manage the array above, and set the default value.

   int         EndRun;
   int         Key;
   char        message[255];
   int         i, j;                // Multi-purpose, always usefull counters.
   int         column1, column2;    // Values fields locations.
   char        log_name[255];       // Log file name.

   int         text_color;          // Colors.
   int         text_hilight_color;
   int         back_color;
   int         graph_back_color;
   int         graph_color;
   int         graph_hilight_color;
   int         graph_x1, graph_y1, graph_x2, graph_y2; // Plotting limits.
   int         graph_vzoom;         // For plotting computations.
   float       graph_hzoom;         // For plotting computations.
   int         graph_point;         // For plotting computations.
   int         last_graph_point;    // For plotting computations.
   int         x1, y1, x2, y2;      // For plotting computations.    
   struct record_t last_record_max; // For maxima marker.
   struct record_t last_record_min; // For minima marker.
   PALETTE     pal; 
   BITMAP      *bmp;
   
// Set parallel port for reading the A/D converter (port previously set to EPP in BIOS).
acquire_port();

//***** Check parameters.
if (argc < 3) {
   printf("Usage: %s <320x200 / 640x480 / 800x600 / 1024x768> <log_file_name> [nowarning]\n", argv[0]);
   exit(1);
   }

//***** Init Allegro.
allegro_init();
install_keyboard(); 
install_timer();

//***** Resolution. 
if (!strcmp(argv[1], "320x200")) {
   x1 = 320; y1 = 200;
   }
else if (!strcmp(argv[1], "640x480")) {
   x1 = 640; y1 = 480;
   }
else if (!strcmp(argv[1], "800x600")) {
   x1 = 800; y1 = 600;
   }
else if (!strcmp(argv[1], "1024x768")) {
   x1 = 1024; y1 = 768;
   }

if (set_gfx_mode(GFX_AUTODETECT, x1, y1, 0, 0) != 0) {
   set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
   allegro_message("Unable to set any graphic mode\n%s\n", allegro_error);
   return 1;
   }
// set_palette(desktop_palette);
// acquire_screen();

//***** Log file name.
strcpy(log_name, argv[2]);

//***** Init markers (max and min solar flux recorded).
last_record_max.solar_flux = 0;
last_record_min.solar_flux = converter_value_max;

//***** Graph windows.
graph_x1          = 1;
graph_y1          = menu_line_max * text_height(font);
graph_x2          = SCREEN_W - 2;
graph_y2          = SCREEN_H - 2;

//***** Colors.
text_color        = makecol(255, 255, 255);
text_hilight_color= makecol(255, 0, 0);
back_color        = makecol(128, 128, 128);
graph_back_color  = makecol(0, 0, 0);
graph_color       = makecol(255, 255, 0);
graph_hilight_color= makecol(0, 255, 0);
graph_vzoom       = 2;
graph_hzoom       = (float) (graph_x2 - graph_x1 - 1) / converter_value_max;

text_mode(-1); // set transparent text.

//***** check call params.
if ((argc >= 4) && (strcmp(argv[3], "nowarning"))) {
   // Display some instructions before running.
   clear_to_color(screen, graph_back_color);
   textout_centre(screen, font, "CAUTION", SCREEN_W/2, SCREEN_H/2-3*text_height(font), text_hilight_color);
   textout_centre(screen, font, "Parallel port must be set to EPP in BIOS.", SCREEN_W/2, SCREEN_H/2-text_height(font), text_color);
   textout_centre(screen, font, "Software must run before plugging A/D converter.", SCREEN_W/2, SCREEN_H/2+2*text_height(font), text_color);
   textout_centre(screen, font, "Call software with /nowarning in command line", SCREEN_W/2, SCREEN_H/2+4*text_height(font), text_hilight_color);
   textout_centre(screen, font, "to make this message disapear.", SCREEN_W/2, SCREEN_H/2+5*text_height(font), text_hilight_color);
   rest(10000);
   }

//***** Main menu.
clear_to_color(screen, back_color);
strcpy(message, "10-12GHz solar radio telescope ");
strcat(message, VERSION);
textout(screen, font, message, 0, 0, text_color);
textout(screen, font, "(C)2005 AstronoScope", 0, text_height(font), text_color);

textout(screen, font, "time.........", 0, 2*text_height(font), text_color);
textout(screen, font, "solar flux...", 0, 3*text_height(font), text_color);
textout(screen, font, "log file [F1]", 0, 4*text_height(font), text_color);
textout(screen, font, "HardCopy [F2]", 0, 5*text_height(font), text_color);
column1 = text_length(font, "XXXXXXXXXXXXX"); // H location for value fields above.

textout(screen, font, "markers     [F3]", column1+text_length(font, "XXXX"), 3*text_height(font), text_color);
textout(screen, font, "Int.(ms) [F4/F5]", column1+text_length(font, "XXXX"), 4*text_height(font), text_color);
textout(screen, font, "vrt zoom    [+-]", column1+text_length(font, "XXXX"), 5*text_height(font), text_color);
column2 = text_length(font, "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"); // H location for value fields above.

// Display numeric horizontal scale.
for (i = 0; i < graph_x2; i+=25) {
   j = i*graph_hzoom;
   sprintf(message, "%d", i);
   textout(screen, font, message, j+2, graph_y1-text_height(font), graph_back_color);
   } 

if (strcmp(argv[1], "320x200")) { // Below 640x480 resolution, hard to display a logo!
   // Display AstronoScope logo.
   bmp = load_bitmap("logo_2005.bmp", NULL);
   if (bmp) {
      //set_palette(pal); // Select the bitmap palette.
      blit(bmp, screen, 0, 0, (SCREEN_W-bmp->w-1), (1), bmp->w, bmp->h);
      x1 = SCREEN_W-bmp->w;
      destroy_bitmap(bmp);
      }
   
   // Display Socit d'Astronomie de Nantes logo.
   bmp = load_bitmap("logo_san.bmp", NULL);
   if (bmp) {
      //set_palette(pal); // Select the bitmap palette.
      blit(bmp, screen, 0, 0, (x1-2-bmp->w), (1), bmp->w, bmp->h);
      destroy_bitmap(bmp);
      }
   }

//***** Init records array.
for (i = 0; i <= record_max; i++) {
   record[i].record_date.da_year = 0;
   record[i].record_date.da_mon  = 0;
   record[i].record_date.da_day  = 0;
   record[i].record_time.ti_hour = 0;
   record[i].record_time.ti_min  = 0;
   record[i].record_time.ti_sec  = 0;
   record[i].record_time.ti_hund = 0;
   record[i].solar_flux = 0;
   }

//***** Init real-time acquisition variables.
LOCK_VARIABLE(record_updated);
LOCK_VARIABLE(record);
LOCK_VARIABLE(f_log);
LOCK_FUNCTION(acquire_solar_flux);
// Start A/D converter acquisition real-time handler.
install_int_ex(acquire_solar_flux, MSEC_TO_TIMER(acquisition_interval[interval]));

//***** Init main loop variables.
EndRun         = FALSE;
record_updated = FALSE;
while (keypressed())
   Key = readkey();
Key = 0;

//***** Main loop (interrupted by real-time acquisition handler).
while (EndRun == FALSE) {
   //***** Trap pressed key, if any.
   release_screen();
   while (keypressed()) {
      Key = readkey();
      }
   acquire_screen();
   
   //***** Action depending on pressed 
   switch (Key) {
      case K_ESCAPE:
         // End of program.
         EndRun = TRUE;
         break;
      case K_F1:
         // Toggle open/close log file.
         if (f_log == NULL)
            open_log(log_name);
         else
            close_log();
         break;
      case K_F2:
         // Hard-copy.
          get_palette(pal); 
          bmp = create_sub_bitmap(screen, 0, 0, SCREEN_W, SCREEN_H); 
          save_bitmap("cyclop.bmp", bmp, pal); 
          destroy_bitmap(bmp); 
         break;
      case K_F3:
         // Reset minima and maxima markers.
         last_record_max.solar_flux = 0;
         last_record_min.solar_flux = converter_value_max;
         // Write a mark onto log file (means that an event occured).
         if (f_log != NULL) {
            fputs("-----------\n", (FILE *) f_log);
            }
         break;
      case K_F4:
         // Decrease acquisition_interval.
         if (interval > interval_min)
            interval--;
         install_int_ex(acquire_solar_flux, MSEC_TO_TIMER(acquisition_interval[interval]));
         break;
      case K_F5:
         // Increase acquisition_interval.
         if (interval < interval_max)
            interval++;
         install_int_ex(acquire_solar_flux, MSEC_TO_TIMER(acquisition_interval[interval]));
         break;
      case K_plus:
         // Increase vertical zoom.
         if (graph_vzoom < 10)
            graph_vzoom++;
         break;
      case K_minus:
         // Decrease vertical zoom.
         if (graph_vzoom > 1)
            graph_vzoom--;
         break;
      case 0:
         break;
      default:
         EndRun = TRUE;
         allegro_message("Key...%d is not managed.", Key);
         break;
      } // end switch key
      
   text_mode(graph_back_color);

   //***** Display last measure.
   sprintf(message, "%04d-%02d-%02d %02d:%02d:%02d:%02d", record[record_max].record_date.da_year, record[record_max].record_date.da_mon, record[record_max].record_date.da_day, record[record_max].record_time.ti_hour, record[record_max].record_time.ti_min, record[record_max].record_time.ti_sec, record[record_max].record_time.ti_hund);
   textout(screen, font, message, column1, 2*text_height(font), text_hilight_color);
   sprintf(message, "%3d", record[record_max].solar_flux);
   textout(screen, font, message, column1, 3*text_height(font), text_hilight_color);

   //***** Display log file status.
   if (f_log == NULL)
      textout(screen, font, "OFF", column1, 4*text_height(font), text_hilight_color);
   else
      textout(screen, font, "ON ", column1, 4*text_height(font), text_hilight_color);

   //***** Display vertical zoom value.
   sprintf(message, "%5d", graph_vzoom);
   textout(screen, font, message, column2, 5*text_height(font), text_hilight_color);
   
   //***** Display acquisition interval.
   sprintf(message, "%5d", acquisition_interval[interval]);
   textout(screen, font, message, column2, 4*text_height(font), text_hilight_color);
   
   //***** Refresh screen if a new record exists.
   if ((record_updated == TRUE) || (Key != 0)) {
      // Following code is executed only if real-time handler is faster.
      record_updated = FALSE;
      
      //***** Display graph.
      rectfill(screen, graph_x1, graph_y1, graph_x2, graph_y2, graph_back_color);
      text_mode(-1);

      // Draw grid:
      // Horizontal division.
      for (i = 0; i < graph_x2; i+=25) {
         j = i*graph_hzoom;
         line (screen, j, graph_y1, j, graph_y2, back_color);
         } 
      // Vertical division.
      for (i = graph_y2; i > graph_y1; i-=25)
         line (screen, graph_x1, i, graph_x2, i, back_color);
      // Plot values.
      for (i = record_max; i > 1; i--) {
         y1 = ploty(record[i], graph_y1, graph_y2, graph_vzoom, i);
         x1 = plotx(record[i], graph_x1, graph_x2, graph_hzoom);
         y2 = ploty(record[i-1], graph_y1, graph_y2, graph_vzoom, i-1);
         x2 = plotx(record[i-1], graph_x1, graph_x2, graph_hzoom);
         line(screen, x1, y1, x2, y2, graph_color);
         // Display values.
         if ((i % (20/graph_vzoom) == 0) &&
             (y1 > graph_y1+text_height(font)) &&
             (y1 < graph_y2-text_height(font))) {
            sprintf(message, "%02d:%02d:%02d:%02d", record[i].record_time.ti_hour, record[i].record_time.ti_min, record[i].record_time.ti_sec, record[i].record_time.ti_hund);
            if (x1 + 1 + text_length(font, "XXXXXXXXXXX") < graph_x2) {
               textout(screen, font, message, x1+text_length(font, "X"), y1-4, text_hilight_color);
               line(screen, x1+2, y1, x1+5, y1, graph_hilight_color);
               }
            else {              
               textout(screen, font, message, x1-text_length(font, "XXXXXXXXXXXX"), y1-4, text_hilight_color);
               line(screen, x1-2, y1, x1-5, y1, graph_hilight_color);
               }
            }
         }
           
      //***** Marker: maxima.
      // Keep maxima in memory.
      if (record[record_max].solar_flux > last_record_max.solar_flux)
         last_record_max = record[record_max];
      // Display maxima.
      y1 = graph_y2-4*text_height(font);
      x1 = plotx(last_record_max, graph_x1, graph_x2, graph_hzoom);
      line(screen, x1, y1-5, x1, y1+15, graph_hilight_color);
      if (x1-5 > graph_x1) // drawing fits graph area.
         triangle (screen, x1-5, y1, x1-1, y1+4, x1-5, y1+8, text_hilight_color);
      else               // reverse drawing
         triangle (screen, x1+5, y1, x1+1, y1+4, x1+5, y1+8, text_hilight_color);
      if (x1-text_length(font, "XXXXXXXXXXX") < graph_x1) // reverse drawing to fit graph area.
         x1 = x1+2+text_length(font, "XXXXXXXXXXX");
      sprintf(message, "%04d-%02d-%02d", last_record_max.record_date.da_year, last_record_max.record_date.da_mon, last_record_max.record_date.da_day);
      textout(screen, font, message, x1-text_length(font, "XXXXXXXXXXX"), y1-2*text_height(font), graph_hilight_color);     
      sprintf(message, "%02d:%02d:%02d:%02d", last_record_max.record_time.ti_hour, last_record_max.record_time.ti_min, last_record_max.record_time.ti_sec, last_record_max.record_time.ti_hund);
      textout(screen, font, message, x1-text_length(font, "XXXXXXXXXXX"), y1-text_height(font), graph_hilight_color);
      sprintf(message, "%3d", last_record_max.solar_flux);
      textout(screen, font, message, x1-text_length(font, "XXXX"), y1, graph_hilight_color);

      //***** Marker: minima.
      // Keep minima in memory.
      if (record[record_max].solar_flux < last_record_min.solar_flux)
         last_record_min = record[record_max];     
      // Display minima.
      y1 = y1-4*text_height(font);
      x1 = plotx(last_record_min, graph_x1, graph_x2, graph_hzoom);
      line(screen, x1, y1-5, x1, y1+15, graph_hilight_color);
      if (x1+5 > graph_x2) // drawing fits graph area.
         triangle (screen, x1-5, y1+8, x1-1, y1+4, x1-5, y1, text_hilight_color);
      else               // reverse drawing
         triangle (screen, x1+5, y1+8, x1+1, y1+4, x1+5, y1, text_hilight_color);
      if (x1+text_length(font, "XXXXXXXXXXX") > graph_x2) // reverse drawing to fit graph area.
         x1 = x1-2-text_length(font, "XXXXXXXXXXX");
      sprintf(message, "%04d-%02d-%02d", last_record_min.record_date.da_year, last_record_min.record_date.da_mon, last_record_min.record_date.da_day);
      textout(screen, font, message, x1+2, y1-2*text_height(font), graph_hilight_color);     
      sprintf(message, "%02d:%02d:%02d:%02d", last_record_min.record_time.ti_hour, last_record_min.record_time.ti_min, last_record_min.record_time.ti_sec, last_record_min.record_time.ti_hund);
      textout(screen, font, message, x1+2, y1-text_height(font), graph_hilight_color);
      sprintf(message, "%3d", last_record_min.solar_flux);
      textout(screen, font, message, x1+2+text_length(font, "X"), y1, graph_hilight_color);
      }
   Key = 0;
   } /* while !EndRun */

//***** Normal termination.
// release_port();               // Left the port set for reading, in order to avoid destruction.
remove_int(acquire_solar_flux);  // Remove the real-time handler.
if (f_log != NULL)               // Close the log file.
   close_log();

return 0;
}

END_OF_MAIN();
